@@ -18,10 +18,10 @@ Currently supported are:
   * svn
   * ssh
 """
-from __future__ import absolute_import
+
 from .fs import URI, FileSystem, BaseUri, RevisionedFileSystem, RevisionedUri
 from .misc import WorkingDirectory
-from exceptions import *
+from .exceptions import *
 
 import logging
 from abl.util import NullHandler
