@@ -13,12 +13,15 @@ class NoSchemeError(PathError):
 class WrongSchemeError(PathError):
     "WrongSchemeError is raised if functionality requires specific schemes"
 
+
 class FileDoesNotExistError(PathError):
     "FileDoesNotExistError is raised, if resource does not exist"
 
+
 class NoDefinedOperationError(PathError):
     "NoDefinedOperationError is raised if method is not supported by backend"
 
+
 class OptionsError(PathError):
     "OptionsError is raised if specific options need to be used"
 
@@ -26,3 +29,7 @@ class OptionsError(PathError):
 class RemoteConnectionTimeout(PathError):
     "Remote connection could not be established"
 
+
+class OperationIsNotSupportedOnPlatform(PathError):
+    "This operation is not supported on this platform"
+
