@@ -1,13 +1,9 @@
-from __future__ import with_statement
-
-
 import atexit
 from collections import defaultdict
-from contextlib import nested
 import fnmatch
 import hashlib
 import os
-from Queue import Queue
+import stat
 import re
 import shutil
 import threading
@@ -18,15 +14,11 @@ from decorator import decorator
 import pkg_resources
 
 from .simpleuri import UriParse, uri_from_parts
-from .exceptions import (
-    NoSchemeError,
-    RemoteConnectionTimeout,
-    FileDoesNotExistError
-    )
+from .exceptions import (NoSchemeError,
+                         FileDoesNotExistError,
+                         OperationIsNotSupportedOnPlatform)
 
 
-#============================================================================
-
 class ConnectionRegistry(object):
     """
     ConnectionRegistry: Singleton for file system backend registry
@@ -123,7 +115,7 @@ class ConnectionRegistry(object):
         """
         if scheme not in self.schemes:
             raise NoSchemeError(
-                'There is no handler registered for "%s"' % scheme
+                'There is no handler registered for "{}" (available: {})'.format(scheme, list(self.schemes.keys()))
                 )
         key = (
             scheme,
@@ -139,24 +131,26 @@ class ConnectionRegistry(object):
             self.create(scheme, key, extras)
         return self.connections[key]
 
+
     def cleanup(self, force=False):
-        import time
         now = time.time()
-        for key, conn in self.connections.items():
+        for key, conn in list(self.connections.items()):
             if ((now - conn.last_used) > self.clean_timeout) or force:
                 try:
                     conn.close()
                 except:
-                    print "### Exception while closing connection %s" % conn
+                    print("### Exception while closing connection %s" % conn)
                     traceback.print_exc()
                 del self.connections[key]
 
+
     def cleaner(self):
         """
         cleaner: method to be run in a thread to check for stale connections.
         """
         while True:
             self.cleanup()
+            now = time.time()
             try:
                 while (time.time() - now) < self.clean_interval:
                     if not self.run_clean_thread:
@@ -165,20 +159,23 @@ class ConnectionRegistry(object):
             except:
                 return
 
+
     def register(self, scheme, factory):
         "register: register factory callable 'factory' with scheme"
 
         self.schemes[scheme] = factory
 
+
     def shutdown(self):
         """
         shutdown: to be run by atexit handler. All open connection are closed.
         """
         self.run_clean_thread = False
         self.cleanup(True)
-        if self.cleaner_thread.isAlive():
+        if self.cleaner_thread.is_alive():
             self.cleaner_thread.join()
 
+
 CONNECTION_REGISTRY = ConnectionRegistry()
 SCHEME_REGISTRY = {}
 
@@ -191,6 +188,7 @@ def normalize_uri(uri, sep='/'):
             uri = '/'+uri[0]+uri[2:]
     return uri
 
+
 def denormalize_path(path, sep='\\'):
     if len(path) > 2 and path[0] == path[2] == '/':
         return path[1]+':'+sep+path[3:].replace('/', sep)
@@ -198,6 +196,25 @@ def denormalize_path(path, sep='\\'):
         return path.replace('/', sep)
 
 
+def remove_dir_recursively(connection, path):
+    for root, dirs, files in connection.walk(path, topdown=False,
+                                             followlinks=False):
+        for fname in files:
+            connection.removefile(root / fname)
+
+        for dname in dirs:
+            d = root / dname
+            if connection.islink(d):
+                connection.removefile(d)
+            else:
+                connection.removedir(d)
+
+    if connection.islink(path):
+        return connection.removefile(path)
+    else:
+        return connection.removedir(path)
+
+
 #============================================================================
 
 @decorator
@@ -210,7 +227,7 @@ def with_connection(func, self, *args, **argd):
     """
     self.connection = self.get_connection()
     self.connection.last_used = time.time()
-
+    self.connection.pre_call_hook(self, func)
     return func(self, *args, **argd)
 
 
@@ -233,6 +250,7 @@ def URI(uri, sep=os.sep, **extras):
         **extras
         )
 
+
 class BaseUri(object):
     """An URI object represents a path, either on the local filesystem or on a
     remote server. On creation, the path is represented by an (more or less
@@ -254,8 +272,8 @@ class BaseUri(object):
         if isinstance(uri, BaseUri):
             self.uri = uri.uri
             self.sep = uri.sep
-            self.parse_result = uri.parse_result
-            self.extras = uri.extras
+            self.parse_result = UriParse(uri.uri)
+            self.extras = uri.extras.copy()
         else:
             if uri.startswith('file://'):
                 uri = uri[7:]
@@ -279,9 +297,11 @@ class BaseUri(object):
             elif 'password' in self.parse_result.query:
                 self.password = self.parse_result.query['password']
 
+
     def get_connection(self):
         return CONNECTION_REGISTRY.get_connection(*self._key(), **self._extras())
 
+
     def _extras(self):
         extras = self.extras.copy()
         if self.scheme not in ('http','https'):
@@ -289,16 +309,19 @@ class BaseUri(object):
 
         return extras
 
+
     def _get_scheme(self):
         if not self._scheme:
             scheme = self.parse_result.scheme
             self._scheme = scheme
         return self._scheme
 
+
     def _set_scheme(self, scheme):
         self._scheme = scheme
         self.parse_result.scheme = scheme
 
+
     scheme = property(_get_scheme, _set_scheme)
 
     @property
@@ -308,6 +331,7 @@ class BaseUri(object):
         except ValueError:
             return None
 
+
     @property
     def path(self):
         path = self.parse_result.path
@@ -316,6 +340,7 @@ class BaseUri(object):
         else:
             return path
 
+
     @property
     def unipath(self):
         pathstr = self.parse_result.path
@@ -324,6 +349,7 @@ class BaseUri(object):
         else:
             return pathstr
 
+
     def _key(self):
         return (
             self.scheme,
@@ -334,6 +360,7 @@ class BaseUri(object):
             self.vpath_connector,
             )
 
+
     def __str__(self):
         return str(self.parse_result)
 
@@ -361,6 +388,10 @@ class BaseUri(object):
         return self.join(other)
 
 
+    def __truediv__(self, other):
+        return self.join(other)
+
+
     def __add__(self, suffix):
         path = self.uri + suffix
         result = self.__class__(
@@ -379,6 +410,7 @@ class BaseUri(object):
             path = path[1:]
         return path.startswith('/')
 
+
     def split(self):
         """
         split: like os.path.split
@@ -405,6 +437,7 @@ class BaseUri(object):
             second.partition('?')[0]
             )
 
+
     def directory(self, level=1):
         """
         @return: the first part of the split method
@@ -420,7 +453,6 @@ class BaseUri(object):
     # os.path-compliance
     dirname = directory
 
-
     def basename(self):
         """
         @return: the second part of the split method
@@ -471,6 +503,7 @@ class BaseUri(object):
             )
         parts = self.parse_result.as_list()
         parts[2] = '/'.join(args)
+
         result = self.__class__(
             uri_from_parts(parts),
             sep=sep,
@@ -481,8 +514,10 @@ class BaseUri(object):
 
         return result
 
+
     @with_connection
-    def copy(self, other, recursive=False, ignore=None):
+    def copy(self, other, recursive=False, ignore=None,
+             followlinks=True):
         """
         copy: copy self to other
 
@@ -490,8 +525,32 @@ class BaseUri(object):
         @param other: the path to copy itself over.
 
         What will really happen depends on the backend.
+
+        Note that file properties are only copied when self and other are
+        located in the same backend, i.e. it is not possible to copy
+        permissions etc. from a memory:// to a file:// based path.
         """
-        return self.connection.copy(self, other, recursive, ignore)
+        return self.connection.copy(self, other,
+                                    recursive=recursive,
+                                    ignore=ignore,
+                                    followlinks=followlinks)
+
+
+    @with_connection
+    def copystat(self, other):
+        """
+        copystat: copy file/folder metadata from self to other
+
+        @type other: URI
+        @param other: the path to copy the metadata to.
+
+        Will copy file permissions, mtime, atime, etc. to path. Note:
+        this ALWAYS follows symlinks. Consider using info with
+        followslinks set to False to read and write the stat info of a
+        symlink.
+        """
+        return self.connection.copystat(self, other)
+
 
     @with_connection
     def move(self, other):
@@ -505,6 +564,7 @@ class BaseUri(object):
         """
         return self.connection.move(self, other)
 
+
     @with_connection
     def remove(self, recursive=False):
         """
@@ -514,27 +574,21 @@ class BaseUri(object):
         the options string contains 'r'. Otherwise, the backends 'removedir'
         method is used.
         """
+        if self.connection.islink(self):
+            return self.connection.removefile(self)
+
         if not self.connection.exists(self):
             raise FileDoesNotExistError(str(self))
+
         if self.connection.isdir(self):
             if recursive:
-                try:
-                    self.connection.rmtree(self)
-                except NotImplementedError:
-                    for root, dirs, files in self.connection.walk(
-                        self,
-                        topdown=False
-                        ):
-                        for fname in files:
-                            self.connection.removefile(root / fname)
-                        for dname in dirs:
-                            self.connection.removedir(root / dname)
-                    return self.connection.removedir(self)
+                return remove_dir_recursively(self.connection, self)
             else:
                 return self.connection.removedir(self)
         elif self.connection.isfile(self):
             return self.connection.removefile(self)
 
+
     @with_connection
     def open(self, options=None, mimetype='application/octet-stream'):
         """
@@ -543,6 +597,7 @@ class BaseUri(object):
         """
         return self.connection.open(self, options, mimetype)
 
+
     @with_connection
     def makedirs(self):
         """
@@ -550,6 +605,7 @@ class BaseUri(object):
         """
         return self.connection.makedirs(self)
 
+
     @with_connection
     def mkdir(self):
         """
@@ -558,6 +614,7 @@ class BaseUri(object):
         """
         return self.connection.mkdir(self)
 
+
     @with_connection
     def exists(self):
         """
@@ -568,6 +625,7 @@ class BaseUri(object):
         """
         return self.connection.exists(self)
 
+
     @with_connection
     def isfile(self):
         """
@@ -578,6 +636,7 @@ class BaseUri(object):
         """
         return self.connection.isfile(self)
 
+
     @with_connection
     def isdir(self):
         """
@@ -589,6 +648,81 @@ class BaseUri(object):
         return self.connection.isdir(self)
 
 
+    @with_connection
+    def isexec(self, mode=stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
+        """
+        isexec:
+
+        @rtype: bool
+        @return: Indicates whether the path points to a file or folder
+        which has the executable flag set.
+
+        Note that on systems which do not support executables flags the
+        result may be unpredicatable.  On Windows the value is determined
+        by the file extension, a *.exe file is executable, a *.sh not.
+        """
+        return self.connection.isexec(self, mode)
+
+
+    @with_connection
+    def set_exec(self, mode):
+        """
+        set_exec:
+
+        @rtype: void
+        @return: Set the executable flag of the receiver to mode (which can
+        be any combination of stat.IXUSR, stat.IXGRP, or stat.IXOTH).
+
+        Note that on systems which do not support executables flags the
+        function may have no effect.
+        """
+        return self.connection.set_exec(self, mode)
+
+
+    @with_connection
+    def islink(self):
+        """
+        islink:
+
+        @rtype: bool
+        @return: Indicates whether the file this path is pointing to is a
+        symlink.
+
+        On systems which don't support symlinks this return alsways False.
+        """
+        return self.connection.islink(self)
+
+
+    @with_connection
+    def readlink(self):
+        """
+        readlink:
+
+        @rtype: URI
+        @return: the content of the symlink this path is pointing to.
+        Throws an exception if this path is not a symlink (check with
+        islink before)
+
+        On systems which don't support symlinks this will always throw.
+        """
+        return self.connection.readlink(self)
+
+
+    @with_connection
+    def symlink(self, link_name):
+        """
+        symlink:
+
+        @rtype: void
+
+        Creates a symbolic link named link_name to the file this path is
+        pointing to.
+
+        On systems which don't support symlinks this will throw.
+        """
+        return self.connection.symlink(self, link_name)
+
+
     @with_connection
     def mtime(self):
         """
@@ -601,7 +735,7 @@ class BaseUri(object):
 
 
     @with_connection
-    def walk(self):
+    def walk(self, topdown=True, followlinks=True):
         """
         walk: walk the filesystem (just like os.walk).
         Use like:
@@ -612,44 +746,32 @@ class BaseUri(object):
 
         root will be an URI object.
         """
-        return self.connection.walk(self)
+        return self.connection.walk(self, topdown=topdown, followlinks=followlinks)
+
 
     @with_connection
-    def relative_walk(self):
-        """
-        similar to "walk", but give as well the
-        directory part relative to the initial directory to walk.
-        Use like:
-
-        path = URI('/some/dir')
-        for root, relative, dirs, files in path.relative_walk():
-            do_something()
-
-        root will be an URI object.
-        relative is a string like "part" or "sub/part"
-        """
-        return self.connection.relative_walk(self)
-
-    @with_connection
-    def listdir(self, recursive=False):
+    def listdir(self):
         """
         listdir: list contents of directory self.
-        if recursive==True, return the content of this
-        directory recursivly. The pathpart of 'self' will
-        not be returned.
+        The pathpart of 'self' will not be returned.
         """
-        return self.connection.listdir(self, recursive)
+        return self.connection.listdir(self)
+
 
     @with_connection
-    def info(self, verbosity=0):
+    def info(self, set_info=None, followlinks=True):
         """
         info: backend info about self (probably not implemented for
-              all backends. The result will be backend specific
+              all backends. The result will be backend specific).
 
+        @param set_info: If not None, this data will be set on self.
+        @param followlinks: If True, symlinks will be followed. If
+            False, the info of the symlink itself will be returned.
+            (Default: True)
         @rtype: Bunch
         @return: backend specific information about self.
         """
-        return self.connection.info(self, verbosity)
+        return self.connection.info(self, set_info=set_info, followlinks=followlinks)
 
 
     @with_connection
@@ -706,14 +828,17 @@ class RevisionedUri(BaseUri):
     def switch(self, branch):
         return self.connection.switch(self, branch)
 
+
     @with_connection
     def update(self, recursive=True, clean=False):
         return self.connection.update(self, recursive, clean)
 
+
     @with_connection
     def log(self, limit=0, **kwargs):
         return self.connection.log(self, limit, **kwargs)
 
+
     @with_connection
     def log_by_time(self, start_time=None, stop_time=None):
         return self.connection.log_by_time(self, start_time, stop_time)
@@ -730,15 +855,13 @@ class FileSystem(object):
 
     scheme = None
 
-    def __init__(
-        self,
-        hostname=None,
-        port=None,
-        username=None,
-        password=None,
-        vpath_connector=None,
-        **extras
-        ):
+    def __init__(self,
+                 hostname=None,
+                 port=None,
+                 username=None,
+                 password=None,
+                 vpath_connector=None,
+                 **extras):
         self.hostname = hostname
         self.port = port
         self.username = username
@@ -749,9 +872,11 @@ class FileSystem(object):
 
         self._initialize()
 
+
     def close(self):
         pass
 
+
     def _path(self, uriobj):
         if isinstance(uriobj, BaseUri):
             return uriobj.path
@@ -759,56 +884,90 @@ class FileSystem(object):
             return uriobj
 
 
+    def pre_call_hook(self, path, func):
+        """
+        Invoked for all @with_connection decorated
+        functions before they get executed.
+        """
+        pass
 
-#-- default implementations -------------------------------------------------
 
-    def copy(self, source, dest, recursive=False, ignore=None):
+    def _copy_link(self, source, dest):
+        assert self.islink(source)
+        if dest.isfile():
+            self.removefile(dest)
+        self.symlink(self.readlink(source), dest)
+
+
+    def copy(self, source, dest, recursive=False, ignore=None,
+             followlinks=True):
         if source.connection is dest.connection and hasattr(self, 'internal_copy'):
             return self.internal_copy(source, dest, recursive, ignore)
 
+        use_same_backend = source.scheme == dest.scheme
 
         if ignore is not None:
             ignore = set(ignore)
         else:
             ignore = set()
+
         if not source.exists():
             raise FileDoesNotExistError(str(source))
+
         if not recursive:
             assert source.isfile()
             if dest.isdir():
                 dest = dest / source.last()
-            with nested(source.open('rb'), dest.open('wb')) as (infs, outfs):
-                shutil.copyfileobj(infs, outfs, 8192)
+
+            if source.islink() and not followlinks:
+                self._copy_link(source, dest)
+            else:
+                with source.open('rb') as infs, dest.open('wb') as outfs:
+                    shutil.copyfileobj(infs, outfs, 8192)
+                if use_same_backend:
+                    self.copystat(source, dest)
         else:
             assert source.isdir()
-            if dest.exists():
+            if dest.isdir():
                 droot = dest / source.last()
             else:
                 droot = dest
-            droot.makedirs()
-            spth = source.path
-            spth_len = len(spth) + 1
-            for root, dirs, files in source.walk():
-                rpth = root.path
-                tojoin = rpth[spth_len:].strip()
-                if tojoin:
-                    dbase = droot / tojoin
-                else:
-                    dbase = droot
-                for folder in dirs[:]:
-                    if folder in ignore:
-                        dirs.remove(folder)
-                        continue
-                    ddir = dbase / folder
-                    ddir.makedirs()
-                for fname in files:
-                    source = root / fname
-                    dest = dbase / fname
-                    with nested(
-                        source.open('rb'),
-                        dest.open('wb')
-                        ) as (infs, outfs):
-                        shutil.copyfileobj(infs, outfs, 8192)
+
+            if source.islink() and not followlinks:
+                self._copy_link(source, droot)
+            else:
+                droot.makedirs()
+                spth = source.path
+                spth_len = len(spth) + 1
+                for root, dirs, files in source.walk(followlinks=followlinks):
+                    rpth = root.path
+                    tojoin = rpth[spth_len:].strip()
+                    if tojoin:
+                        dbase = droot / tojoin
+                    else:
+                        dbase = droot
+
+                    for folder in dirs[:]:
+                        if folder in ignore:
+                            dirs.remove(folder)
+                            continue
+                        ddir = dbase / folder
+                        srcp = root / folder
+                        if srcp.islink() and not followlinks:
+                            self._copy_link(srcp, ddir)
+                        else:
+                            ddir.makedirs()
+
+                    for fname in files:
+                        srcf = root / fname
+                        destf = dbase / fname
+                        if srcf.islink() and not followlinks:
+                            self._copy_link(srcf, destf)
+                        else:
+                            with srcf.open('rb') as infs, destf.open('wb') as outfs:
+                                shutil.copyfileobj(infs, outfs, 8192)
+                            if use_same_backend:
+                                self.copystat(srcf, destf)
 
 
     def makedirs(self, path):
@@ -835,33 +994,11 @@ class FileSystem(object):
             source.remove('r')
 
 
-    def relative_walk(self, top, relative="", topdown=True):
-        names = self.listdir(top)
+    def walk(self, top, topdown=True, followlinks=True):
+        names = sorted(self.listdir(top))
 
-        dirs, nondirs = [], []
-        for name in names:
-            if self.isdir(top / name):
-                dirs.append(name)
-            else:
-                nondirs.append(name)
-
-        if topdown:
-            yield top, relative, dirs, nondirs
-        for name in dirs:
-            path = top / name
-            if not relative:
-                relpart = name
-            else:
-                relpart = relative + '/%s' % name
-            for x in self.relative_walk(path, relpart, topdown):
-                yield x
-        if not topdown:
-            yield top, relative, dirs, nondirs
-
-    def walk(self, top, topdown=True):
-        names = self.listdir(top)
-
-        dirs, nondirs = [], []
+        dirs = []
+        nondirs = []
         for name in names:
             if self.isdir(top / name):
                 dirs.append(name)
@@ -872,8 +1009,9 @@ class FileSystem(object):
             yield top, dirs, nondirs
         for name in dirs:
             path = top / name
-            for x in self.walk(path, topdown):
-                yield x
+            if not path.islink() or followlinks:
+                for x in self.walk(path, topdown=topdown, followlinks=followlinks):
+                    yield x
         if not topdown:
             yield top, dirs, nondirs
 
@@ -882,21 +1020,18 @@ class FileSystem(object):
         # TODO-std: is this working with separators?
         res = []
         for f in path.listdir():
-            f = path / f
-            if fnmatch.fnmatch(self._path(f), pattern):
-                res.append(f)
+            if fnmatch.fnmatch(f, pattern):
+                res.append(path / f)
         return res
 
 
-#-- overwritable methods ----------------------------------------------------
-
     def _initialize(self):
         raise NotImplementedError
 
     def open(self, path, options, mimetype):
         raise NotImplementedError
 
-    def listdir(self, path, recursive=False):
+    def listdir(self, path):
         raise NotImplementedError
 
     def removefile(self, path):
@@ -905,9 +1040,6 @@ class FileSystem(object):
     def removedir(self, path):
         raise NotImplementedError
 
-    def rmtree(self, path):
-        raise NotImplementedError
-
     def mkdir(self, path):
         raise NotImplementedError
 
@@ -920,25 +1052,64 @@ class FileSystem(object):
     def isdir(self, path):
         raise NotImplementedError
 
-    def info(self,  path, verbosity=0):
+    def isexec(self, path, mode):
         raise NotImplementedError
 
+    def set_exec(self, path, mode):
+        raise NotImplementedError
+
+
+    def supports_symlinks(self):
+        raise NotImplementedError
+
+
+    def islink(self, path):
+        if not self.supports_symlinks():
+            return False
+        raise NotImplementedError
+
+
+    def readlink(self, path):
+        if not self.supports_symlinks():
+            raise OperationIsNotSupportedOnPlatform
+        raise NotImplementedError
+
+
+    def symlink(self, target, link_name):
+        if not self.supports_symlinks():
+            raise OperationIsNotSupportedOnPlatform
+        raise NotImplementedError
+
+
+    def info(self,  path, set_info=None, followlinks=True):
+        raise NotImplementedError
+
+
     def sync(self, source, dest, options):
         raise NotImplementedError
 
+
     def mtime(self, path):
         raise NotImplementedError
 
+
+    def copystat(self, path, other):
+        raise NotImplementedError
+
+
 class RevisionedFileSystem(FileSystem):
     def switch(self, revision):
         raise NotImplementedError
 
+
     def update(self, path, recursive=False, clean=False):
         raise NotImplementedError
 
+
     def log(self, path, limit=0):
         raise NotImplementedError
 
+
     def log_by_time(self, path, start_time=None, stop_time=None):
         raise NotImplementedError
 
@@ -946,8 +1117,8 @@ class RevisionedFileSystem(FileSystem):
 for entrypoint in pkg_resources.iter_entry_points('abl.vpath.plugins'):
     try:
         plugin_class = entrypoint.load()
-    except Exception, exp:
-        print "Could not load entrypoint", entrypoint
+    except Exception as exp:
+        print("Could not load entrypoint", entrypoint)
         traceback.print_exc()
         continue
     CONNECTION_REGISTRY.register(plugin_class.scheme, plugin_class)
