@@ -2,22 +2,31 @@
 # (C) 2008 Ableton AG
 # Author: Stephan Diehl <stephan.diehl@ableton.com>
 #******************************************************************************
-from __future__ import with_statement, absolute_import
 
 import datetime
 import logging
+import os
 import shutil
 import stat
-import subprocess
+import sys
 
-from .fs import FileSystem, BaseUri, denormalize_path
-from .exceptions import FileDoesNotExistError
-from .os_abstraction import os
+from functools import partial
+
+from .fs import FileSystem, BaseUri, denormalize_path, URI
+from .exceptions import OperationIsNotSupportedOnPlatform
 
 from abl.util import Bunch, LockFile
 
+
 LOGGER = logging.getLogger(__name__)
-#----------------------------------------------------------------------------
+
+
+def lchmod(filename, mode):
+    if sys.version_info[0] >= 3:
+        return os.chmod(filename, mode, follow_symlinks=True)
+    else:
+        return os.lchmod(filename, mode)
+
 
 class LocalFileSystemUri(BaseUri):
     def __str__(self):
@@ -31,6 +40,7 @@ class LocalFileSystemUri(BaseUri):
         else:
             return super(LocalFileSystemUri, self).path
 
+
 class LocalFileSystem(FileSystem):
     scheme = 'file'
     uri = LocalFileSystemUri
@@ -39,15 +49,19 @@ class LocalFileSystem(FileSystem):
         pass
 
 
-    def info(self, unc, set_info=None):
+    def info(self, unc, set_info=None, followlinks=True):
         p = self._path(unc)
 
+        use_link_functions = self.islink(unc) and not followlinks
+        stat_func = os.lstat if use_link_functions else os.stat
+        chmod_func = lchmod if use_link_functions else os.chmod
+
         if set_info is not None:
             if "mode" in set_info:
-                os.chmod(p, set_info["mode"])
+                chmod_func(p, set_info["mode"])
             return
 
-        stats = os.stat(p)
+        stats = stat_func(p)
         ctime = stats[stat.ST_CTIME]
         mtime = stats[stat.ST_MTIME]
         atime = stats[stat.ST_ATIME]
@@ -67,21 +81,19 @@ class LocalFileSystem(FileSystem):
         else:
             return open(self._path(unc))
 
-    def listdir(self, unc, recursive=False):
-        return os.listdir(self._path(unc))
+    def listdir(self, unc):
+        return sorted(os.listdir(self._path(unc)))
 
 
     def removefile(self, unc):
         pth = self._path(unc)
+        if sys.platform == 'win32':
+            os.chmod(pth, stat.S_IWUSR)
         return os.unlink(pth)
 
 
-    def rmtree(self, unc):
-        pth = self._path(unc)
-        return shutil.rmtree(pth)
-
-
     def removedir(self, unc):
+        assert not unc.islink()
         pth = self._path(unc)
         return os.rmdir(pth)
 
@@ -91,15 +103,29 @@ class LocalFileSystem(FileSystem):
         if path:
             return os.mkdir(path)
 
+
     def exists(self, unc):
         return os.path.exists(self._path(unc))
 
+
     def isfile(self, unc):
         return os.path.isfile(self._path(unc))
 
+
     def isdir(self, unc):
         return os.path.isdir(self._path(unc))
 
+
+    def isexec(self, unc, mode_mask):
+        return self.info(unc).mode & mode_mask != 0
+
+
+    def set_exec(self, unc, mode):
+        mask = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
+        new_mode = (self.info(unc).mode & ~mask) | mode
+        self.info(unc, dict(mode=new_mode))
+
+
     def move(self, source, dest):
         """
         the semantic should be like unix 'mv' command.
@@ -131,3 +157,29 @@ class LocalFileSystem(FileSystem):
     def mtime(self, path):
         return self.info(path).mtime
 
+
+    def copystat(self, source, dest):
+        shutil.copystat(source.path, dest.path)
+
+
+    def supports_symlinks(self):
+        return sys.platform != 'win32'
+
+
+    def islink(self, path):
+        return self.supports_symlinks() and os.path.islink(self._path(path))
+
+
+    def readlink(self, path):
+        if not self.supports_symlinks():
+            raise OperationIsNotSupportedOnPlatform
+        return URI(os.readlink(self._path(path)))
+
+
+    def symlink(self, target, link_name):
+        if not self.supports_symlinks():
+            raise OperationIsNotSupportedOnPlatform
+        return os.symlink(self._path(target), self._path(link_name))
+
+    def dump(self, outf, no_binary=False):
+        pass
